# -*- coding: utf-8 -*-
"""
Created on 2021/3/4 11:26 下午
---------
@summary:
---------
@author: Boris
@email: boris_liu@foxmail.com
"""

from feapder import Request, Response


def test_selector():
    request = Request("https://www.baidu.com?a=1&b=2", data={}, params=None)
    response = request.get_response()
    print(response)

    print(response.xpath("//a/@href"))
    print(response.css("a::attr(href)"))
    print(response.css("a::attr(href)").extract_first())

    content = response.re("<a.*?href='(.*?)'")
    print(content)


def test_from_text():
    text = """    <script src="./lib/docsify/lib/docsify.min.js"></script>
        <script src="./lib/docsify/lib/plugins/ga.js"></script>
        <script src="./lib/docsify/lib/plugins/search.js"></script>
        <script src="./lib/docsify-copy-code/dist/docsify-copy-code.min.js"></script>
        <script src="./lib/prismjs/components/prism-bash.js"></script>
        <script src="./lib/prismjs/components/prism-java.js"></script>
        <script src="./lib/prismjs/components/prism-sql.js"></script>
        <script src="./lib/prismjs/components/prism-yaml.js"></script>
        <script src="./lib/prismjs/components/prism-python.js"></script>
        <script src="//cdn.jsdelivr.net/npm/docsify/lib/plugins/zoom-image.min.js"></script>"""

    resp = Response.from_text(text=text, url="http://feapder.com/#/README")
    print(resp.text)
    print(resp)
    print(resp.xpath("//script"))



# -*- coding: utf-8 -*-
"""爬虫配置文件"""
import os

# redis 表名
# 任务表模版
TAB_REQUESTS = "{redis_key}:z_requests"
# 任务失败模板
TAB_FAILED_REQUESTS = "{redis_key}:z_failed_requests"
# 数据保存失败模板
TAB_FAILED_ITEMS = "{redis_key}:s_failed_items"
# 爬虫状态表模版
TAB_SPIDER_STATUS = "{redis_key}:h_spider_status"
# 用户池
TAB_USER_POOL = "{redis_key}:h_{user_type}_pool"

# MYSQL
MYSQL_IP = os.getenv("MYSQL_IP")
MYSQL_PORT = int(os.getenv("MYSQL_PORT", 3306))
MYSQL_DB = os.getenv("MYSQL_DB")
MYSQL_USER_NAME = os.getenv("MYSQL_USER_NAME")
MYSQL_USER_PASS = os.getenv("MYSQL_USER_PASS")

# MONGODB
MONGO_IP = os.getenv("MONGO_IP", "localhost")
MONGO_PORT = int(os.getenv("MONGO_PORT", 27017))
MONGO_DB = os.getenv("MONGO_DB")
MONGO_USER_NAME = os.getenv("MONGO_USER_NAME")
MONGO_USER_PASS = os.getenv("MONGO_USER_PASS")

# REDIS
# ip:port 多个可写为列表或者逗号隔开 如 ip1:port1,ip2:port2 或 ["ip1:port1", "ip2:port2"]
REDISDB_IP_PORTS = os.getenv("REDISDB_IP_PORTS")
REDISDB_USER_PASS = os.getenv("REDISDB_USER_PASS")
REDISDB_DB = int(os.getenv("REDISDB_DB", 0))
# 适用于redis哨兵模式
REDISDB_SERVICE_NAME = os.getenv("REDISDB_SERVICE_NAME")

# 数据入库的pipeline，可自定义，默认MysqlPipeline
ITEM_PIPELINES = [
    "feapder.pipelines.mysql_pipeline.MysqlPipeline",
    # "feapder.pipelines.mongo_pipeline.MongoPipeline",
]
EXPORT_DATA_MAX_FAILED_TIMES = 10  # 导出数据时最大的失败次数，包括保存和更新，超过这个次数报警
EXPORT_DATA_MAX_RETRY_TIMES = 10  # 导出数据时最大的重试次数，包括保存和更新，超过这个次数则放弃重试

# 爬虫相关
# COLLECTOR
COLLECTOR_TASK_COUNT = 10  # 每次获取任务数量

# SPIDER
SPIDER_THREAD_COUNT = 1  # 爬虫并发数
# 下载时间间隔 单位秒。 支持随机 如 SPIDER_SLEEP_TIME = [2, 5] 则间隔为 2~5秒之间的随机数，包含2和5
SPIDER_SLEEP_TIME = 0
SPIDER_MAX_RETRY_TIMES = 100  # 每个请求最大重试次数
# 是否主动执行添加 设置为False 需要手动调用start_monitor_task，适用于多进程情况下
SPIDER_AUTO_START_REQUESTS = True
KEEP_ALIVE = False  # 爬虫是否常驻

# 浏览器渲染
WEBDRIVER = dict(
    pool_size=1,  # 浏览器的数量
    load_images=True,  # 是否加载图片
    user_agent=None,  # 字符串 或 无参函数，返回值为user_agent
    proxy=None,  # xxx.xxx.xxx.xxx:xxxx 或 无参函数，返回值为代理地址
    headless=False,  # 是否为无头浏览器
    driver_type="CHROME",  # CHROME、PHANTOMJS、FIREFOX
    timeout=30,  # 请求超时时间
    window_size=(1024, 800),  # 窗口大小
    executable_path=None,  # 浏览器路径，默认为默认路径
    render_time=0,  # 渲染时长，即打开网页等待指定时间后再获取源码
    custom_argument=["--ignore-certificate-errors"],  # 自定义浏览器渲染参数
    xhr_url_regexes=None,  # 拦截xhr接口，支持正则，数组类型
    auto_install_driver=False,  # 自动下载浏览器驱动 支持chrome 和 firefox
)

# 爬虫启动时，重新抓取失败的requests
RETRY_FAILED_REQUESTS = False
# 保存失败的request
SAVE_FAILED_REQUEST = True
# request防丢机制。（指定的REQUEST_LOST_TIMEOUT时间内request还没做完，会重新下发 重做）
REQUEST_LOST_TIMEOUT = 600  # 10分钟
# request网络请求超时时间
REQUEST_TIMEOUT = 22  # 等待服务器响应的超时时间，浮点数，或(connect timeout, read timeout)元组
# item在内存队列中最大缓存数量
ITEM_MAX_CACHED_COUNT = 5000
# item每批入库的最大数量
ITEM_UPLOAD_BATCH_MAX_SIZE = 1000
# item入库时间间隔
ITEM_UPLOAD_INTERVAL = 1

# 下载缓存 利用redis缓存，但由于内存大小限制，所以建议仅供开发调试代码时使用，防止每次debug都需要网络请求
RESPONSE_CACHED_ENABLE = False  # 是否启用下载缓存 成本高的数据或容易变需求的数据，建议设置为True
RESPONSE_CACHED_EXPIRE_TIME = 3600  # 缓存时间 秒
RESPONSE_CACHED_USED = False  # 是否使用缓存 补采数据时可设置为True

# redis 存放item与request的根目录
REDIS_KEY = ""
# 爬虫启动时删除的key，类型: 元组/bool/string。 支持正则; 常用于清空任务队列，否则重启时会断点续爬
DELETE_KEYS = []

# 设置代理
PROXY_EXTRACT_API = None  # 代理提取API ，返回的代理分割符为\r\n
PROXY_ENABLE = True

# 随机headers
RANDOM_HEADERS = True
# UserAgent类型 支持 'chrome', 'opera', 'firefox', 'internetexplorer', 'safari'，'mobile' 若不指定则随机类型
USER_AGENT_TYPE = "chrome"
# 默认使用的浏览器头 RANDOM_HEADERS=True时不生效
DEFAULT_USERAGENT = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/73.0.3683.103 Safari/537.36"
# requests 使用session
USE_SESSION = False

# 下载
DOWNLOADER = "feapder.network.downloader.RequestsDownloader"
SESSION_DOWNLOADER = "feapder.network.downloader.RequestsSessionDownloader"
MAKE_ABSOLUTE_LINKS = True  # 自动转成绝对连接

# 去重
ITEM_FILTER_ENABLE = False  # item 去重
ITEM_FILTER_SETTING = dict(
    filter_type=1  # 永久去重（BloomFilter） = 1 、内存去重（MemoryFilter） = 2、 临时去重（ExpireFilter）= 3
)
REQUEST_FILTER_ENABLE = False  # request 去重
REQUEST_FILTER_SETTING = dict(
    filter_type=3,  # 永久去重（BloomFilter） = 1 、内存去重（MemoryFilter） = 2、 临时去重（ExpireFilter）= 3
    expire_time=2592000,  # 过期时间1个月
)

# 报警 支持钉钉、飞书、企业微信、邮件
# 钉钉报警
DINGDING_WARNING_URL = ""  # 钉钉机器人api
DINGDING_WARNING_PHONE = ""  # 报警人 支持列表，可指定多个
DINGDING_WARNING_ALL = False  # 是否提示所有人， 默认为False
# 飞书报警
# https://open.feishu.cn/document/ukTMukTMukTM/ucTM5YjL3ETO24yNxkjN#e1cdee9f
FEISHU_WARNING_URL = ""  # 飞书机器人api
FEISHU_WARNING_USER = None  # 报警人 {"open_id":"ou_xxxxx", "name":"xxxx"} 或 [{"open_id":"ou_xxxxx", "name":"xxxx"}]
FEISHU_WARNING_ALL = False  # 是否提示所有人， 默认为False
# 邮件报警
EMAIL_SENDER = ""  # 发件人
EMAIL_PASSWORD = ""  # 授权码
EMAIL_RECEIVER = ""  # 收件人 支持列表，可指定多个
EMAIL_SMTPSERVER = "smtp.163.com"  # 邮件服务器 默认为163邮箱
# 企业微信报警
WECHAT_WARNING_URL = ""  # 企业微信机器人api
WECHAT_WARNING_PHONE = ""  # 报警人 将会在群内@此人, 支持列表，可指定多人
WECHAT_WARNING_ALL = False  # 是否提示所有人， 默认为False
# 时间间隔
WARNING_INTERVAL = 3600  # 相同报警的报警时间间隔，防止刷屏; 0表示不去重
WARNING_LEVEL = "DEBUG"  # 报警级别， DEBUG / INFO / ERROR
WARNING_FAILED_COUNT = 1000  # 任务失败数 超过WARNING_FAILED_COUNT则报警
WARNING_CHECK_TASK_COUNT_INTERVAL = 1200  # 检查已做任务数量的时间间隔，若两次时间间隔之间，任务数无变化则报警

# 日志
LOG_NAME = os.path.basename(os.getcwd())
LOG_PATH = "log/%s.log" % LOG_NAME  # log存储路径
LOG_LEVEL = "DEBUG"
LOG_COLOR = True  # 是否带有颜色
LOG_IS_WRITE_TO_CONSOLE = True  # 是否打印到控制台
LOG_IS_WRITE_TO_FILE = False  # 是否写文件
LOG_MODE = "w"  # 写文件的模式
LOG_MAX_BYTES = 10 * 1024 * 1024  # 每个日志文件的最大字节数
LOG_BACKUP_COUNT = 20  # 日志文件保留数量
LOG_ENCODING = "utf8"  # 日志文件编码
# 是否详细的打印异常
PRINT_EXCEPTION_DETAILS = True
# 设置不带颜色的日志格式
LOG_FORMAT = "%(threadName)s|%(asctime)s|%(filename)s|%(funcName)s|line:%(lineno)d|%(levelname)s| %(message)s"
# 设置带有颜色的日志格式
os.environ["LOGURU_FORMAT"] = (
    "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | "
    "<level>{level: <8}</level> | "
    "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>line:{line}</cyan> | <level>{message}</level>"
)
OTHERS_LOG_LEVAL = "ERROR"  # 第三方库的log等级

# 打点监控 influxdb 配置
INFLUXDB_HOST = os.getenv("INFLUXDB_HOST", "localhost")
INFLUXDB_PORT = int(os.getenv("INFLUXDB_PORT", 8086))
INFLUXDB_UDP_PORT = int(os.getenv("INFLUXDB_UDP_PORT", 8089))
INFLUXDB_USER = os.getenv("INFLUXDB_USER")
INFLUXDB_PASSWORD = os.getenv("INFLUXDB_PASSWORD")
INFLUXDB_DATABASE = os.getenv("INFLUXDB_DB")
# 监控数据存储的表名，爬虫管理系统上会以task_id命名
INFLUXDB_MEASUREMENT = "task_" + os.getenv("TASK_ID") if os.getenv("TASK_ID") else None
# 打点监控其他参数，若这里也配置了influxdb的参数, 则会覆盖外面的配置
METRICS_OTHER_ARGS = dict(retention_policy_duration="180d", emit_interval=60)

############# 导入用户自定义的setting #############
try:
    from setting import *

    # 兼容老版本的配置
    KEEP_ALIVE = not AUTO_STOP_WHEN_SPIDER_DONE
except:
    pass



# -*- coding: utf-8 -*-
"""
Created on 2021/4/8 1:06 下午
---------
@summary:
---------
@author: Boris
@email: boris_liu@foxmail.com
"""

from feapder.utils.perfect_dict import PerfectDict


task_key = ["id", "url"]
task = [1, "http://www.badu.com"]
task = Task(_dict=dict(zip(task_key, task)), _values=task)

task = Task(id=1, url="http://www.badu.com")
task = Task({"id":"1", "url":"http://www.badu.com"})

print(task)
id, url = task
print(id, url)
print(task[0], task[1])
print(task.id, task.url)
print(task["id"], task["url"])
print(task.get("id"), task.get("url"))



# -*- coding: utf-8 -*-
"""
Created on 2020/4/22 12:05 AM
---------
@summary: 基于内存队列的爬虫，不支持分布式
---------
@author: Boris
@email: boris_liu@foxmail.com
"""

from threading import Thread

import feapder.setting as setting
import feapder.utils.tools as tools
from feapder.buffer.item_buffer import ItemBuffer
from feapder.core.base_parser import BaseParser
from feapder.core.parser_control import AirSpiderParserControl
from feapder.db.memory_db import MemoryDB
from feapder.network.request import Request
from feapder.utils.log import log
from feapder.utils import metrics


class AirSpider(BaseParser, Thread):
    __custom_setting__ = {}

    def __init__(self, thread_count=None):
        """
        基于内存队列的爬虫，不支持分布式
        :param thread_count: 线程数
        """
        super(AirSpider, self).__init__()

        for key, value in self.__class__.__custom_setting__.items():
            setattr(setting, key, value)

        if thread_count:
            setattr(setting, "SPIDER_THREAD_COUNT", thread_count)
        self._thread_count = setting.SPIDER_THREAD_COUNT

        self._memory_db = MemoryDB()
        self._parser_controls = []
        self._item_buffer = ItemBuffer(redis_key="air_spider")

        metrics.init(**setting.METRICS_OTHER_ARGS)

    def distribute_task(self):
        for request in self.start_requests():
            if not isinstance(request, Request):
                raise ValueError("仅支持 yield Request")

            request.parser_name = request.parser_name or self.name
            self._memory_db.add(request)

    def all_thread_is_done(self):
        for i in range(3):  # 降低偶然性, 因为各个环节不是并发的，很有可能当时状态为假，但检测下一条时该状态为真。一次检测很有可能遇到这种偶然性
            # 检测 parser_control 状态
            for parser_control in self._parser_controls:
                if not parser_control.is_not_task():
                    return False

            # 检测 任务队列 状态
            if not self._memory_db.empty():
                return False

            # 检测 item_buffer 状态
            if (
                self._item_buffer.get_items_count() > 0
                or self._item_buffer.is_adding_to_db()
            ):
                return False

            tools.delay_time(1)

        return True

    def run(self):
        self.start_callback()

        for i in range(self._thread_count):
            parser_control = AirSpiderParserControl(self._memory_db, self._item_buffer)
            parser_control.add_parser(self)
            parser_control.start()
            self._parser_controls.append(parser_control)

        self._item_buffer.start()

        self.distribute_task()

        while True:
            try:
                if self.all_thread_is_done():
                    # 停止 parser_controls
                    for parser_control in self._parser_controls:
                        parser_control.stop()

                    # 关闭item_buffer
                    self._item_buffer.stop()

                    # 关闭webdirver
                    if Request.webdriver_pool:
                        Request.webdriver_pool.close()

                    log.info("无任务，爬虫结束")
                    break

            except Exception as e:
                log.exception(e)

            tools.delay_time(1)  # 1秒钟检查一次爬虫状态

        self.end_callback()
        # 为了线程可重复start
        self._started.clear()
        # 关闭打点
        metrics.close()

    def join(self, timeout=None):
        """
        重写线程的join
        """
        if not self._started.is_set():
            return

        super().join()



# -*- coding: utf-8 -*-
"""
Created on 2018-12-13 21:08
---------
@summary:
---------
@author: Boris
@email: boris_liu@foxmail.com
"""

import copy
from typing import Any, List, Union, Optional, Tuple, Callable

from feapder.utils.tools import get_md5
from .bloomfilter import BloomFilter, ScalableBloomFilter
from .expirefilter import ExpireFilter


class Dedup:
    BloomFilter = 1
    MemoryFilter = 2
    ExpireFilter = 3

    def __init__(self, filter_type: int = BloomFilter, to_md5: bool = True, **kwargs):
        """
        去重过滤器 集成BloomFilter、MemoryFilter、ExpireFilter
        Args:
            filter_type: 过滤器类型 BloomFilter
            name: 过滤器名称 该名称会默认以dedup作为前缀 dedup:expire_set:[name]/dedup:bloomfilter:[name]。 默认ExpireFilter name=过期时间; BloomFilter name=dedup:bloomfilter:bloomfilter
            absolute_name: 过滤器绝对名称 不会加dedup前缀，当此值不为空时name参数无效
            expire_time: ExpireFilter的过期时间 单位为秒，其他两种过滤器不用指定
            error_rate: BloomFilter/MemoryFilter的误判率 默认为0.00001
            to_md5: 去重前是否将数据转为MD5，默认是
            redis_url: redis://[[username]:[password]]@localhost:6379/0
                       BloomFilter 与 ExpireFilter 使用
                       默认会读取setting中的redis配置，若无setting，则需要专递redis_url
            initial_capacity: 单个布隆过滤器去重容量 默认100000000，当布隆过滤器容量满时会扩展下一个布隆过滤器
            error_rate：布隆过滤器的误判率 默认0.00001
            **kwargs:
        """

        if filter_type == Dedup.ExpireFilter:
            try:
                expire_time = kwargs["expire_time"]
            except:
                raise ValueError("需传参数 expire_time")

            name = kwargs.get("absolute_name") or "dedup:expire_set:%s" % kwargs.get(
                "name", expire_time
            )
            expire_time_record_key = "dedup:expire_set:expire_time"

            self.dedup = ExpireFilter(
                name=name,
                expire_time=expire_time,
                expire_time_record_key=expire_time_record_key,
                redis_url=kwargs.get("redis_url"),
            )

        else:
            initial_capacity = kwargs.get("initial_capacity", 100000000)
            error_rate = kwargs.get("error_rate", 0.00001)
            name = kwargs.get("absolute_name") or "dedup:bloomfilter:" + kwargs.get(
                "name", "bloomfilter"
            )
            if filter_type == Dedup.BloomFilter:
                self.dedup = ScalableBloomFilter(
                    name=name,
                    initial_capacity=initial_capacity,
                    error_rate=error_rate,
                    bitarray_type=ScalableBloomFilter.BASE_REDIS,
                    redis_url=kwargs.get("redis_url"),
                )
            elif filter_type == Dedup.MemoryFilter:
                self.dedup = ScalableBloomFilter(
                    name=name,
                    initial_capacity=initial_capacity,
                    error_rate=error_rate,
                    bitarray_type=ScalableBloomFilter.BASE_MEMORY,
                )
            else:
                raise ValueError(
                    "filter_type 类型错误，仅支持 Dedup.BloomFilter、Dedup.MemoryFilter、Dedup.ExpireFilter"
                )

        self._to_md5 = to_md5

    def __repr__(self):
        return str(self.dedup)

    def _deal_datas(self, datas):
        if self._to_md5:
            if isinstance(datas, list):
                keys = [get_md5(data) for data in datas]
            else:
                keys = get_md5(datas)
        else:
            keys = copy.deepcopy(datas)

        return keys

    def add(
        self, datas: Union[List[Any], Any], skip_check: bool = False
    ) -> Union[List[Any], Any]:
        """
        添加数据
        @param datas: list / 单个值
        @param skip_check: 是否直接添加，不检查是否存在 适用于bloomfilter，加快add速度
        @return: list / 单个值 (如果数据已存在 返回 0 否则返回 1, 可以理解为是否添加成功)
        """

        keys = self._deal_datas(datas)
        is_added = self.dedup.add(keys, skip_check)

        return is_added

    def get(self, datas: Union[List[Any], Any]) -> Union[List[Any], Any]:
        """
        检查数据是否存在
        @param datas: list / 单个值
        @return: list / 单个值 （存在返回1 不存在返回0)
        """
        keys = self._deal_datas(datas)
        is_exists = self.dedup.get(keys)

        return is_exists

    def filter_exist_data(
        self,
        datas: List[Any],
        *,
        datas_fingerprints: Optional[List] = None,
        callback: Callable[[Any], None] = None
    ) -> Union[Tuple[List[Any], List[Any]], List[Any]]:
        """
        过滤掉已存在的数据
        *** 直接修改原来的数据 使用完此方法后 datas, datas_fingerprints 里面的值为去重后的数据
        @param datas_fingerprints: 数据的唯一指纹 列表
        @param datas: 数据 列表
        @param callback: 数据已存在时的回调 callback(data)
        @return: None
        """

        is_exists = self.get(datas_fingerprints or datas)

        dedup_datas = []

        if datas_fingerprints:
            dedup_datas_fingerprints = []
            while is_exists:
                data = datas.pop(0)
                is_exist = is_exists.pop(0)
                data_fingerprint = datas_fingerprints.pop(0)

                if not is_exist:
                    dedup_datas.append(data)
                    dedup_datas_fingerprints.append(data_fingerprint)
                else:
                    if callback:
                        callback(data)

            datas_fingerprints.extend(dedup_datas_fingerprints)
            datas.extend(dedup_datas)
            return datas, datas_fingerprints

        else:
            while is_exists:
                data = datas.pop(0)
                is_exist = is_exists.pop(0)

                if not is_exist:
                    dedup_datas.append(data)
                else:
                    if callback:
                        callback(data)

            datas.extend(dedup_datas)
            return datas



# -*- coding: utf-8 -*-
"""
Created on 2021-02-08 16:09:47
---------
@summary:
---------
@author: Boris
"""

import feapder
from items import *


class TestSpider(feapder.BatchSpider):
    # def init_task(self):
    #     pass

    def start_requests(self, task):
        # task 为在任务表中取出的每一条任务
        id, url = task  # id， url为所取的字段，main函数中指定的
        yield feapder.Request(url, task_id=id)

    def parse(self, request, response):
        title = response.xpath('//title/text()').extract_first()  # 取标题
        item = spider_data_item.SpiderDataItem()  # 声明一个item
        item.title = title  # 给item属性赋值
        yield item  # 返回item， item会自动批量入库
        yield self.update_task_batch(request.task_id, 1) # 更新任务状态为1

    def exception_request(self, request, response):
        """
        @summary: 请求或者parser里解析出异常的request
        ---------
        @param request:
        @param response:
        ---------
        @result: request / callback / None (返回值必须可迭代)
        """

        pass

    def failed_request(self, request, response):
        """
        @summary: 超过最大重试次数的request
        ---------
        @param request:
        ---------
        @result: request / item / callback / None (返回值必须可迭代)
        """

        yield request
        yield self.update_task_batch(request.task_id, -1) # 更新任务状态为-1


